/*------------------------------------------------------------------------
  File: soundey.i
  Description: Routines similar to the "Soundex" algorithms
  Input Parameters: 
      <none>
  Output Parameters:
      <none>
  Author: S.E. Southwell
  Copyright: FreeFrameWork 2001
  Created: 10/24/01
  Last Modified:
  
  Soundex:               Soundey:
  0 = aehiouwy          0 = aehiouwy
  1 = bfpv              1 = bp 
  2 = cgjkqsxz          2 = ckqx
  3 = dt                3 = dt
  4 = l                 4 = l
  5 = mn                5 = mn
  6 = r                 6 = r
                        7 = fv
                        8 = gj
                                    9 = sz
------------------------------------------------------------------------*/

&SCOPED-DEFINE tranlist "01230780082455012693070209"
&SCOPED-DEFINE letrlist "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
&SCOPED-DEFINE multicon     "CK,PF,PH,UGH,TION,NG,DT,CE,CI,AX,EX,IX,OX,UX,YX,^XA,^XI,^XE,^XO,^XU,FTH"
&SCOPED-DEFINE multicontran "K,F,F,UF,SHUN,N,T,SE,SI,AKS,EKS,IKS,OKS,UKS,IKS,^ZA,^ZI,^XE,^XO,^XU,TH"
&SCOPED-DEFINE endings "ATES$,ETES$,ITES$,OTES$,UTES$,YTES$,ANE$,ENE$,INE$,ONE$,UNE$,YNE$,AKE$,EKE$,IKE$,OKE$,UKE$,YKE$,AKES$,EKES$,IKES$,OKES$,UKES$,YKES$"
&SCOPED-DEFINE endingtrans "ATZ$,ETZ$,ITZ$,OTZ$,UTZ$,YTZ$,AN$,EN$,IN$,ON$,UN$,YN$,AK$,EK$,IK$,OK$,UK$,YK$,AKS$,EKS$,IKS$,OKS$,UKS$,YKS$"
&SCOPED-DEFINE vowels "A,E,H,I,O,U,W,Y"

&SCOPED-DEFINE numbers "1ST,2ND,3RD,4TH,5TH,6TH,7TH,8TH,9TH,1,2,3,4,5,6,7,8,9,0"
&SCOPED-DEFINE spokenumbers "FURST,SEKUNT,THURD,FORTH,FITH,SIXTH,SEVUNTH,ATH,NINTH,ON,TU,THRE,FOR,FIVE,SIKS,SEVEN,AT,NINE,ZERO"
&SCOPED-DEFINE ignorechars "'"
&SCOPED-DEFINE terminators '~"~~`!^()/|~{~}[]' + CHR(92) 
&SCOPED-DEFINE digits "1,2,3,4,5,6,7,8,9,0"
&SCOPED-DEFINE beforedigits ",.-"

FUNCTION wordize RETURNS CHAR(
 INPUT myString AS CHAR
):
    DEFINE VAR charcount AS INTEGER NO-UNDO.
    DEFINE VAR count2 AS INTEGER NO-UNDO.
    /* Split into words */

    /* Ignore certain chars */
    DO charcount = 1 TO LENGTH({&ignorechars}):
        ASSIGN myString = REPLACE(myString,SUBSTRING({&ignorechars},charcount,1),"").
    END. /*charcount*/

    /*Mark up letters, temporarily, so we can identify them*/
    DO charcount = 1 TO LENGTH({&LETRLIST}):
        ASSIGN myString = REPLACE(myString,SUBSTRING({&LETRLIST},charcount,1),"'" + SUBSTRING({&LETRLIST},charcount,1) + "'").
    END. /*charcount*/

    /*Now change abbreviations like U.S.A and O.K. to USA and OK*/
    ASSIGN myString = REPLACE(myString,"'.'","").
    
    /*Unmark the letters*/
    ASSIGN myString = REPLACE(myString,"'","").
    
    
    /* Use some chars to break up words */
    DO charcount = 1 TO LENGTH({&TERMINATORS}):
        ASSIGN myString = REPLACE(myString,SUBSTRING({&TERMINATORS},charcount,1)," ").
    END. /*charcount*/

    /* Mark up the digits, temporarily, so we can identify them */
    DO charcount = 1 TO NUM-ENTRIES({&DIGITS}):
        ASSIGN myString = REPLACE(myString,ENTRY(charcount,{&DIGITS}),"'" + ENTRY(charcount,{&DIGITS})).
    END. /*charcount*/

    /* Now break words EXCEPT where the beforedigits chars are used before digits*/
    DO charcount = 1 TO (LENGTH(mystring) - 1):
        DO count2 = 1 TO LENGTH({&BEFOREDIGITS}):
            IF SUBSTRING(mystring,charcount,1) = SUBSTRING({&BEFOREDIGITS},count2,1)
             AND SUBSTRING(myString,charcount + 1,1) NE "'" 
             THEN ASSIGN SUBSTRING(mystring,charcount,1) = " ".
        END. /*count2*/
    END. /*charcount*/
    
    /*Unmark the digits*/
    ASSIGN myString = REPLACE(myString,"'","").
    
    
    /* Combine multiple whitespace */
    DO count2 = 1 to 31:
        ASSIGN myString = REPLACE(myString,chr(count2)," ").
    END.
    ASSIGN myString = REPLACE(myString,"~n"," ").
    ASSIGN myString = REPLACE(myString,"~t"," ").
    ASSIGN myString = REPLACE(myString,"~r"," ").
    ASSIGN myString = REPLACE(myString,chr(10), " ").
    ASSIGN myString = REPLACE(myString,chr(11), " ").
    ASSIGN myString = REPLACE(myString,chr(12), " ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = TRIM(myString).
    
    /* Now mark the word boundaries, so we can replace specific prefixes and suffixes*/
    ASSIGN myString = "^" + myString + "$".
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString,"  "," ").
    ASSIGN myString = REPLACE(myString," ","$ ^").
    RETURN myString.

END FUNCTION. /*wordize*/

FUNCTION unBoundarize RETURNS CHAR(
 INPUT myString AS CHAR
):
    ASSIGN myString = TRIM(myString,"$^").
    ASSIGN myString = REPLACE(myString,"$ ^"," ").
    ASSIGN myString = TRIM(REPLACE(myString,"  "," ")).
    RETURN myString.
END FUNCTION. /*UNBOUNDARIZE*/

FUNCTION preToken RETURNS CHAR(
 INPUT myString AS CHAR
):
    DEFINE VAR charcount AS INTEGER NO-UNDO.

    ASSIGN myString = CAPS(myString).
    ASSIGN myString = Wordize(myString).
    /* Turn digits into phonetic sounds */
    DO charcount = 1 TO NUM-ENTRIES({&NUMBERS}):
        ASSIGN myString = REPLACE(myString,ENTRY(charcount,{&NUMBERS}),ENTRY(charcount,{&SPOKENUMBERS})).
    END.



    /* Pre-process multi-consonant sounds into equivalent single-consonant sounds */
    DO charcount = 1 TO NUM-ENTRIES({&MULTICON}):
        ASSIGN myString = REPLACE(myString,ENTRY(charcount,{&MULTICON}),ENTRY(charcount,{&MULTICONTRAN})).
    END.
        
    /* Pre-process word-endings to eliminate silent vowels and such*/
    DO charcount = 1 TO NUM-ENTRIES({&ENDINGS}):
        ASSIGN myString = REPLACE(myString,ENTRY(charcount,{&ENDINGS}) + "$",ENTRY(charcount,{&ENDINGTRANS}) + "$").
    END.
        

 RETURN myString.

END FUNCTION.


FUNCTION toSoundey RETURNS CHAR(
 INPUT myString AS CHAR
):
    DEFINE VAR myReturn AS CHAR NO-UNDO.
    DEFINE VAR charcount AS INTEGER NO-UNDO.
    DEFINE VAR thisChar AS CHAR NO-UNDO.
    DEFINE VAR lastChar AS CHAR NO-UNDO.
    
    


    ASSIGN myString = preToken(myString).
    DO charcount = 1 TO LENGTH(myString):
        ASSIGN thisChar = SUBSTRING(mystring,charcount,1).
        IF charcount > 1 THEN ASSIGN lastChar = SUBSTRING(mystring,charcount - 1,1).
        IF ASC(thisChar) > 64 
         AND ASC(thisChar) < 91 
         THEN DO:
            IF (CAN-DO({&vowels},lastChar))
             AND (CAN-DO({&vowels},thisChar)) 
             THEN NEXT.
            ELSE ASSIGN myReturn = myReturn + substring({&tranlist},(ASC(thischar) - 64),1).
        END. /*was a letter*/
        ELSE ASSIGN myReturn = myReturn + thisChar.
    END. /*each char*/
    
    ASSIGN myReturn = unBoundarize(myReturn).
    ASSIGN myReturn = REPLACE(myreturn,"11","1").
    ASSIGN myReturn = REPLACE(myreturn,"22","2").
    ASSIGN myReturn = REPLACE(myreturn,"33","3").
    ASSIGN myReturn = REPLACE(myreturn,"44","4").
    ASSIGN myReturn = REPLACE(myreturn,"55","5").
    ASSIGN myReturn = REPLACE(myreturn,"66","6").
    ASSIGN myReturn = REPLACE(myreturn,"77","7").
    ASSIGN myReturn = REPLACE(myreturn,"88","8").
    ASSIGN myReturn = REPLACE(myreturn,"99","9").
    RETURN myReturn.
END FUNCTION. /*toSoundey*/

FUNCTION simScore RETURNS DECIMAL(
 INPUT string1 AS CHAR,
 INPUT string2 AS CHAR
):
/*Compare two strings for similarity. 1 = exact same, 0 = dissimilar*/
    DEFINE VAR s1 AS CHAR NO-UNDO.
    DEFINE VAR s2 AS CHAR NO-UNDO.
    DEFINE VAR myreturn AS DECIMAL INIT 1.0 NO-UNDO.
    DEFINE VAR slicecoeff AS DECIMAL NO-UNDO.
    DEFINE VAR lengthcoeff AS DECIMAL NO-UNDO.
    DEFINE VAR wordsimcoeff AS DECIMAL NO-UNDO.
    DEFINE VAR soundsimcoeff AS DECIMAL INIT 1.0 NO-UNDO.
    DEFINE VAR thiswordsimcoeff AS DECIMAL INIT 1.0 NO-UNDO.
    DEFINE VAR thissoundsimcoeff AS DECIMAL INIT 1.0 NO-UNDO.
    DEFINE VAR charcount AS INTEGER NO-UNDO.
    DEFINE VAR movecount AS INTEGER NO-UNDO.
    DEFINE VAR thisreturn AS DECIMAL NO-UNDO.
    DEFINE VAR gapfiller  AS CHAR NO-UNDO.
    
    ASSIGN
     string1 = TRIM(string1)
     string2 = TRIM(string2)
    .
    
    IF string1 = string2 THEN RETURN 1.0.
    
    IF length(string1) >= length(string2) THEN ASSIGN
     s1 = toSoundey(string1)
     s2 = toSoundey(string2)
    . 
    ELSE ASSIGN
     s1 = toSoundey(string2)
     s2 = toSoundey(string1)
    .
    
    ASSIGN
     slicecoeff = 1 / MAX(LENGTH(string1),LENGTH(string2))
     lengthcoeff = 1 - ABS(slicecoeff * (LENGTH(string1) - LENGTH(string2)))
    . 
    ASSIGN myreturn = lengthcoeff * myreturn.
    
    /* Now eliminate double-letters to help with comparison */
    DO charcount = 1 to LENGTH({&LETRLIST}):
        ASSIGN
         string1 = REPLACE(string1,SUBSTRING({&LETRLIST},charcount,1) + SUBSTRING({&LETRLIST},charcount,1), SUBSTRING({&LETRLIST},charcount,1))
         string2 = REPLACE(string2,SUBSTRING({&LETRLIST},charcount,1) + SUBSTRING({&LETRLIST},charcount,1), SUBSTRING({&LETRLIST},charcount,1))
        .
    END.
    
    IF LENGTH(string1) > LENGTH(string2) THEN
    DO movecount = 0 to LENGTH(string1) - LENGTH(string2):
        ASSIGN thiswordsimcoeff = 1.0.
        DO charcount = 1 to LENGTH(string2):
            IF SUBSTRING(string1,charcount + movecount,1) NE SUBSTRING(string2,charcount,1) 
             THEN DO:
                ASSIGN thiswordsimcoeff = thiswordsimcoeff - slicecoeff.
                /* add back in if shifted letter matches: */
                IF charcount > 1 AND SUBSTRING(string1,charcount + movecount - 1,1) = SUBSTRING(string2,charcount,1) 
                 THEN ASSIGN thiswordsimcoeff = thiswordsimcoeff + (slicecoeff * .8).
                ELSE IF charcount < LENGTH(string1) AND SUBSTRING(string1,charcount + movecount + 1,1) = SUBSTRING(string2,charcount,1) 
                 THEN ASSIGN thiswordsimcoeff = thiswordsimcoeff + (slicecoeff * .8).
            END.
        END.
        IF thiswordsimcoeff > wordsimcoeff THEN ASSIGN wordsimcoeff = thiswordsimcoeff.
    END.
    IF LENGTH(string2) > LENGTH(string1) THEN
    DO movecount = 0 to LENGTH(string2) - LENGTH(string1):
        ASSIGN thiswordsimcoeff = 1.0.
        DO charcount = 1 to LENGTH(string2):
            IF SUBSTRING(string2,charcount + movecount,1) NE SUBSTRING(string1,charcount,1) 
             THEN DO:
                ASSIGN thiswordsimcoeff = thiswordsimcoeff - slicecoeff.
                /* add back in if shifted letter matches: */
                IF charcount > 1 AND SUBSTRING(string2,charcount + movecount - 1,1) = SUBSTRING(string1,charcount,1) 
                 THEN ASSIGN thiswordsimcoeff = thiswordsimcoeff + (slicecoeff * .8).
                ELSE IF charcount < LENGTH(string2) AND SUBSTRING(string2,charcount + movecount + 1,1) = SUBSTRING(string1,charcount,1) 
                 THEN ASSIGN thiswordsimcoeff = thiswordsimcoeff + (slicecoeff * .8).
            END.
        END.
        IF thiswordsimcoeff > wordsimcoeff THEN ASSIGN wordsimcoeff = thiswordsimcoeff.
    END.
    IF LENGTH(string2) = LENGTH(string1) THEN DO:
        ASSIGN wordsimcoeff = 1.0.
        DO charcount = 1 to LENGTH(string2):
            IF SUBSTRING(string2,charcount,1) NE SUBSTRING(string1,charcount,1) 
             THEN DO:
                ASSIGN wordsimcoeff = wordsimcoeff - slicecoeff.
                /* add back in if shifted letter matches: */
                IF charcount > 1 AND SUBSTRING(string2,charcount - 1,1) = SUBSTRING(string1,charcount,1) 
                 THEN ASSIGN wordsimcoeff = wordsimcoeff + (slicecoeff * .8).
                ELSE IF charcount < LENGTH(string2) AND SUBSTRING(string2,charcount + 1,1) = SUBSTRING(string1,charcount,1) 
                 THEN ASSIGN wordsimcoeff = wordsimcoeff + (slicecoeff * .8).
            END. /*matching against neighboring characters*/
        END.
    END.
    ASSIGN myreturn = wordsimcoeff * myreturn.

/*****************************************************************
*  NOW COMPARE THE SOUNDEY'S 
******************************************************************/    
    ASSIGN soundsimcoeff = 0.
    DO movecount = 0 to ABS(LENGTH(s1) - LENGTH(s2)):
        ASSIGN thissoundsimcoeff = 1.
        DO charcount = 1 to MIN(LENGTH(s1),LENGTH(s2)):
            IF SUBSTRING(s1,charcount + movecount,1) NE SUBSTRING(s2,charcount,1) 
             THEN DO:
                ASSIGN thissoundsimcoeff = thissoundsimcoeff - slicecoeff.
                /* Add back in if neighboring characters match */
                IF LENGTH(s1) >= LENGTH(s2) THEN DO:
                    IF charcount > 1 AND substring(s1,charcount + movecount - 1,1) = SUBSTRING(s2,charcount,1)
                     THEN ASSIGN thissoundsimcoeff = thissoundsimcoeff + (slicecoeff * .8).
                    ELSE IF charcount < MIN(LENGTH(S1),LENGTH(S2)) AND substring(s1,charcount + movecount + 1,1) = SUBSTRING(s2,charcount,1)
                     THEN ASSIGN thissoundsimcoeff = thissoundsimcoeff + (slicecoeff * .8).
                END.
            END. /*MATCHING NEIGHBORING CHARACTERS*/
        END.
        IF thissoundsimcoeff > soundsimcoeff THEN ASSIGN soundsimcoeff = thisSoundsimcoeff.
    END.  /*movecount*/
    ASSIGN myreturn = soundsimcoeff * myreturn.
    
    /* Now, if the soundey's match, assess a "bonus" of half the distance to the goal*/
    IF soundsimcoeff = 1 THEN ASSIGN myReturn = (myreturn + 1) / 2.
    
/*****************************************************************
*  HANDLE LENGTH DIFFERENCES BY BREAKING IN HALF AND MATCHING THE HALVES
******************************************************************/    

    /* IF lengths are unequal, try running this whole thing again with things shifted*/
    DO charcount = 1 TO ABS(LENGTH(string1) - LENGTH(string2)):
        ASSIGN gapfiller = gapfiller + "-".
    END.
    IF LENGTH(string1) > LENGTH(string2) THEN DO:
        DO charcount = 0 TO LENGTH(string2):
            IF charcount = 0 THEN ASSIGN thisreturn = lengthcoeff * simScore(string1,gapfiller + string2).
            ELSE IF charcount = LENGTH(string2) THEN ASSIGN thisreturn = lengthcoeff * simScore(string1,string2 + gapfiller).
            ELSE ASSIGN thisreturn = lengthcoeff * simScore(string1,SUBSTRING(string2,1,charcount) + gapfiller + SUBSTRING(string2,charcount + 1,-1)).
            IF thisreturn > myreturn THEN ASSIGN myreturn = thisreturn.
        END.
    END.
    IF LENGTH(string2) > LENGTH(string1) THEN DO:
        DO charcount = 0 TO LENGTH(string1):
            IF charcount = 0 THEN ASSIGN thisreturn = lengthcoeff * simScore(string2,gapfiller + string1).
            ELSE IF charcount = LENGTH(string1) THEN ASSIGN thisreturn = lengthcoeff * simScore(string2,string1 + gapfiller).
            ELSE ASSIGN thisreturn = lengthcoeff * simScore(string2,SUBSTRING(string1,1,charcount) + gapfiller + SUBSTRING(string1,charcount + 1,-1)).
            IF thisreturn > myreturn THEN ASSIGN myreturn = thisreturn.
        END.
    END.
/* queue-message("steve",string1 + " - " + string2 + " - " + string(soundsimcoeff) + " - " + string(wordsimcoeff) + " - " + string(lengthcoeff) + " - " + string(myreturn)).*/
    RETURN myreturn.
    
END FUNCTION. /*simScore*/

PROCEDURE wordMatch:
    DEFINE INPUT PARAMETER p-bigstring AS CHAR NO-UNDO.
    DEFINE INPUT PARAMETER p-littlestring AS CHAR NO-UNDO.
    DEFINE OUTPUT PARAMETER p-bestmatch AS CHAR NO-UNDO.
    DEFINE OUTPUT PARAMETER p-bestcoeff AS DECIMAL NO-UNDO.
    DEFINE VAR wordcount AS INTEGER NO-UNDO.
    DEFINE VAR thismatch AS DECIMAL NO-UNDO.

    IF p-bigstring + p-littlestring = ? THEN RETURN.
    IF p-bigstring = "" or p-littlestring = "" THEN RETURN.
    
    /* Need to split bigstring into multiple words */
    ASSIGN p-bigstring = wordize(p-bigstring).
    ASSIGN p-bigstring = unboundarize(p-bigstring).
   
    DO wordcount = 1 TO NUM-ENTRIES(p-bigstring," "):
        ASSIGN thismatch = simScore(ENTRY(wordcount,p-bigstring," "),p-littlestring).
        IF thismatch > p-bestcoeff THEN DO:
            ASSIGN
             p-bestmatch = ENTRY(wordcount,p-bigstring," ")
             p-bestcoeff = thismatch
            .
        END. /*found a candidate*/
        IF thismatch = 1 THEN RETURN.
    END. 
END PROCEDURE.

/*
{lib/soundey.i}
DEFINE VAR myword AS CHAR NO-UNDO.
DEFINE VAR charcount AS INTEGER NO-UNDO.
DEFINE VAR wordlist AS CHAR NO-UNDO.

DEFINE TEMP-TABLE tt-word 
 FIELD wordname AS CHAR
 FIELD wordscore AS DECIMAL
 FIELD hits AS INTEGER
 FIELD relevance AS DECIMAL
 INDEX wordname is unique wordname
 INDEX wordscore is primary wordscore
 INDEX relevance relevance
.

ASSIGN myword = "star".
ASSIGN wordlist = myword.

DO CHARCOUNT = 1 TO LENGTH(myword) - 1:
    ASSIGN wordlist = wordlist + " " + substring(myword,1,charcount) + " " + substring(myword,charcount + 1, -1).
END.
DO CHARCOUNT = 1 to NUM-ENTRIES(wordlist," "):
  IF LENGTH(ENTRY(charcount,wordlist," ")) < min(5,length(myword)) 
   THEN ASSIGN ENTRY(charcount,wordlist," ") = "".
END.
ASSIGN wordlist = replace(wordlist,"  "," ").

{&out} wordlist.
DO charcount = 1 TO NUM-ENTRIES(wordlist," "):
for each word where word.soundfraglist contains toSoundey(ENTRY(charcount,wordlist," ")):
display word.
put stream webstream control null (0).
    find tt-word 
     where tt-word.wordname = word.word
     no-error.
    if not avail tt-word then do:
        create tt-word.
        assign
         tt-word.wordscore = simScore(myword,word.word)  
         tt-word.wordname = word.word
        .
    end.
    ASSIGN
     tt-word.hits = tt-word.hits + 1
     tt-word.relevance = tt-word.hits * tt-word.wordscore
    .
end.
END.

for each tt-word by relevance descending:
    display tt-word.wordname format "x(20)" tt-word.wordscore tt-word.hits tt-word.relevance. 
end.
*/
